/////////////////////////////////////////////////////////////////////////////
//
// texture.cpp : implementation of the cTexture class
//
/////////////////////////////////////////////////////////////////////////////


#include "texture.h"
#include <math.h>
#include <assert.h>
#include <stdio.h>
#include "glext.h"

//BEGIN IMPORTANT SWITCHES========================

const bool cTexture::USEMIPMAP = true;
	/*	The USEMIPMAP switch happens to be implemented as a cTexture static vairable
	instead of as a #define, like most of our switches.
		If I set USEMIPMAP = true, then a texture will actually store a variety of
	textures of different resolution, and pick the best one according to the size of
	the target onscreen pixel swatch.
		One might expect USEMIPMAP = true to make the textures run a bit faster or look a
	bit nicer, but I haven't seen any real difference in the limited number of tests I've
	done.
		You might want to have USEMIPMAP = false, simply because if you don't
	use mipmaps, then the RAM size of your texture objects will be less.  But if you use
	a lot of textures, do test the true option.
		Do note that USEMIPMAP = false requires that the resource bitmaps fed into the
	texture-maker have both	dimensions be exact powers of two. We have some code to 
	handle the case where you disobey the powers-of-two-edged source bitmap injuction
	, but	you really SHOULD always use
	powers-of-two-edged source bitmaps. */
#define SMOOTHTEXTURES 
	/* SMOOTHTEXTURES only matters if you don't have USEMIPMAP true.
	If we define SMOOTHTEXTURES in, our texture maps are made by averaging pixel 
	neighbors. This is slower during the construction of the texture, but once the texture
	is created, the speed of usign it is  the same.  So since the construction only happens
	for one brief moment (unless you are continually brininging textures in and out
	of existence) SMOOTHETEXTURES is preferred  */
const int cTexture::MAXEDGESIZE = 1024; /* MAXEDGESIZE should be a power of two. You need
	to increase MAXEDGESIZE here if you plan to use a bigger kind of soruce bitmap. */
//#define GLUSCALETEXTUREINFO
	/*  I thought we needed this becausein order to use a texture in OpenGL,
it should be based on a bitmap to have both edges be powers of two. Actually,
it's not that important because when you use gluBuild2DMipmaps, this 
method will in fact fix the size if it has to.
	But it works ok and doesn't to any harm in case we ever switch off 
mipmaps.
*/

//END IMPORTANT SWITCHES========================

//TEST SWITCHES and an OBSOLETE SWITCH
#define ALLOWALPHATRANSPARENCY //TEST PURPOSES ONLY
	/* If we comment in ALLOWALPHATRANSPARENCY then a cSpriteIcon can be drawn with its
background transparent by using alpha blending.  If we comment out ALLOWALPHATRANSPARENCY,
the sprite icon are drawn with a solid background, that looks a bit clunky. The switch takes
place inside the cTexture constructor code. The purpose of the ALLOWALPHATRANSPARENCY is to
see if using this effect slows the program down much.  The fact seems to be using the
transparency has no substantial speed penalty.  The program will run slow whenever you
have a visually large cSpriteIcon in 3D, whether or not you are using the transparency
effect.  You need to not let the user get too close to a cSpriteIcon
critter, so that you aren't doing a texture map onto a very large number of screen 
pixels. Concievably you can partly alleviate the slowness of large texture areas by
by basing the texture on a large bitmap and making cTexture::USEMIPMAP be true.  */


// cTexture construction/destruction////////////////////////////////////////////////////


cTexture::cTexture(const char *skinfilename, bool transparent):
_transparent(transparent),
_usemipmap(cTexture::USEMIPMAP),
_textureID(0),
_isvalid(false)
{
	cTextureInfo *ptextureinfo = NULL;
	ptextureinfo = new cTextureInfo(skinfilename);
	if (ptextureinfo)
		_isvalid = loadTextureInfo(ptextureinfo);
	delete ptextureinfo;
}

cTexture::cTexture(cTextureInfo *ptextureinfo, bool transparent):
_transparent(transparent),
_usemipmap(cTexture::USEMIPMAP),
_textureID(0),
_isvalid(false)
{
	if (ptextureinfo)
		_isvalid = loadTextureInfo(ptextureinfo);
}

void cTexture::select()
{
	int currenttextureid;
	::glGetIntegerv(GL_TEXTURE_BINDING_2D,
			//The OpenGL doc mistakenly lists this
			//constant as GL_TEXTURE_2D_BINDING
		 &currenttextureid);
	if ((unsigned int)currenttextureid != _textureID)
   		::glBindTexture(GL_TEXTURE_2D, _textureID);
}

bool cTexture::loadTextureInfo(cTextureInfo *ptextureinfo)
{
	if (!ptextureinfo->isValid())
		return false;
	if (_textureID) //Rare case of reloading into same texture.
		::glDeleteTextures(1, &_textureID);

	if (ptextureinfo->textureType == BMP)
	{
//As of texture4.zip April 27, I use alpha in bmp files.
//		_pixeldataformat = GL_RGB;
//		_componentsperpixel = 3;
		_pixeldataformat = GL_RGBA;
		_componentsperpixel = 4;
	}
	else if(ptextureinfo->textureType == PCX)
	{
		_pixeldataformat = GL_RGBA;
		_componentsperpixel = 4;
	}
	else //Unknown file type
		return false;
	int imagecx = _cx = ptextureinfo->_width;
	int imagecy = _cy = ptextureinfo->_height;
	_imageaspect = (double) _cx / (double) _cy;/* By default we set the aspect to match the aspect
		of the  image we loaded, before doing the makePowersOfTwo call. Even though we are going
		to clamp the texture to powers of two, if we apply it to a rectangle with proportions
		matching _imageaspect, the image will appear unstretched; the two stretches cancel out. */

#ifdef GLUSCALETEXTUREINFO
	makePowersOfTwo(_cx, _cy); 
	if (_cx != imagecx || _cy != imagecy) //Need to rescale
	{
		unsigned long arraysize =
			 _componentsperpixel * (unsigned long)_cx * (unsigned long)_cy;
		unsigned char * pdatascaled = new unsigned char[arraysize];
		assert(pdatascaled);
		::glPixelStorei(GL_UNPACK_ALIGNMENT, 1); //We're not padding at ends of lines when we write textures.
		::glPixelStorei(GL_PACK_ALIGNMENT, 1); //We're not padding at ends of lines when we read textures.
		::gluScaleImage(_pixeldataformat, //data format
			imagecx, imagecy, GL_UNSIGNED_BYTE, ptextureinfo->pdata(),
				//settings of source image
			_cx, _cy, GL_UNSIGNED_BYTE, pdatascaled);
				//settings of target image
		const GLubyte* errorstring = gluErrorString(glGetError());
		ptextureinfo->resetData(_cx, _cy, pdatascaled);
	}		
#endif //GLUSCALETEXTUREINFO
	if (_transparent)
	{
		unsigned char *pdata = ptextureinfo->pdata(); //Just a shortcut name.
		unsigned int colorref, transparentcolorref;
		unsigned int rval, gval, bval;
		rval = (unsigned int)(pdata[0]);
		gval = (unsigned int)(pdata[1]);
		bval = (unsigned int)(pdata[2]);
		transparentcolorref = (rval<<24) | (gval<<16) | (bval<<8);
		/* We will treat the color found in the upper left corner as transparent. */
		long unsigned int index;
		for (long unsigned int j = 0; j < (unsigned long int)_cy; j++)
		{
			for (long unsigned int i = 0; i < (unsigned long int)_cx; i++)
			{
				index = 4*(j*_cx + i); //Index of the (i,j) pixel of 4 bytes				rval = (unsigned int)(pdata[index]);
				rval = (unsigned int)(pdata[index]);
				gval = (unsigned int)(pdata[index+1]);
				bval = (unsigned int)(pdata[index+2]);
				colorref = (rval<<24) | (gval<<16) | (bval<<8);
				pdata[index+3] = ((colorref == transparentcolorref) ? 0 : 255); //if then else
				/* Set the alpha to minimum on "background" pixels, and to maximum
					otherwise.  These values get normalized to 0.0 and 1.0 in
					the fragment operations. */
			}
		}
	}
	_maketexture(ptextureinfo->pdata());
	return true;
}

cTexture::~cTexture()
{
	::glDeleteTextures(1, &_textureID);
}

//cTexture helper methods----------------------------------------------
void cTexture::_maketexture(GLubyte *pdata)
{
/* usemipmap uses more memory, makes a nicer looking
		images across a range of size scales.  It may be slower. */
	::glGenTextures(1, &_textureID); /* Get an unused name for a texture
		object, which is something a bit like a display list.  That is,
		it's a kind of handle which will enable us to rapidly reaccess a
		texture once we've first loaded it. */
	::glBindTexture(GL_TEXTURE_2D, _textureID); /* Makes this texture current
		so that all the following calls in _maketexture() affect it. */
	::glPixelStorei(GL_UNPACK_ALIGNMENT, 1); //We're not padding at ends of lines to have even byte cound.
	::glPixelStorei(GL_PACK_ALIGNMENT, 1); //We're not padding at ends of lines to have even byte cound.

	::glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT); //GL_CLAMP less likely to make artifacts.
	::glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	if (_usemipmap)
	{
		::glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,
						GL_LINEAR);
		::glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,
				GL_LINEAR);
	    ::gluBuild2DMipmaps(GL_TEXTURE_2D, 
			_componentsperpixel, //Is bytes per pixel here, 3 or 4.
			_cx, _cy, //Width and height of the source bitmap
			_pixeldataformat,/* Internal GL format to use for pixel component intensities
				can also be viewed as number of color components in texture:
				1,2,3,4 or GL_LUMINANCE, GL_LUMINANCE_ALPHA, GL_RGB, GL_RGBA. */
			GL_UNSIGNED_BYTE, //Type of the intensity values
			pdata); //Source array of intensities of pixel components
		int errorcode = ::glGetError();
		const GLubyte* errorstring = ::gluErrorString(errorcode);
	}
	else //No usemipmap means use only use one texture
	{
#ifndef SMOOTHTEXTURES
	//Use faster size filters.
		::glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,
			 GL_NEAREST);
		::glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
			 GL_NEAREST);
#else //SMOOTHTEXTURES
	//Slower.  LINEAR means average a 2x2 neighborhood.
		::glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,
			 GL_LINEAR);
		::glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
			 GL_LINEAR);
#endif //SMOOTHTEXTURES
		::glTexImage2D( GL_TEXTURE_2D, 
			0, //Level of detail currently being defined, starting with 0.
			_componentsperpixel, //Is bytes per pixel here, 3 or 4.
			_cx, _cy, //Width and height of the source bitmap
			0, //Width of border, if any, in pixels.
			_pixeldataformat,//Internal GL format to use for pixel data pattern
			GL_UNSIGNED_BYTE, //Type of the intensity values
			pdata); //Source array of intensities of pixel components
	}
	::glBindTexture(GL_TEXTURE_2D, 0); //Close off the call to glBindTexture started above.
}




//========================cTextureInfo  File Reading Stuff============

#define BITMAP_ID 0x4D42		// the universal bitmap ID

// LoadTexture()
// desc: loads a texture's information given the skinfilename
// 
bool cTextureInfo::LoadTexture(const char *skinfilename)
{
	// get extension from skinfilename
	char * extStr = strchr(skinfilename, '.');
	extStr++;
	// set the texture type based on extension of skinfilename
	if ((strcmp(extStr, "BMP") == 0) || (strcmp(extStr, "bmp") == 0))
		_loaded = LoadBMPTexture(skinfilename);
	else if ((strcmp(extStr, "PCX") == 0) || (strcmp(extStr, "pcx") == 0) )
		_loaded = LoadPCXTexture(skinfilename);
	return _loaded;
}

cTextureInfo::~cTextureInfo()
{
	if (_pdata)
	{
		delete [] _pdata;
		_pdata = NULL;
	}
	if (_ppalette)
	{
		delete [] _ppalette;
		_ppalette = NULL;
	}
}

// LoadBMPTexture()
// desc: loads a texture of the BMP format
bool cTextureInfo::LoadBMPTexture(const char *skinfilename)
{
	BITMAPINFOHEADER headerinfo;		// BMP header

// store BMP data in texture
	_pdata = LoadBitmapFile(skinfilename, &headerinfo);
	/* Note that unlike PCX, we don't bother scaling the rectangle edges to powers of two. */
	if (_pdata == NULL)
		return false;
// store texture information
	_width = headerinfo.biWidth;
	_height = headerinfo.biHeight;
	_ppalette = NULL;
	textureType = BMP;
	return true;
}

// LoadBitmapFile
// desc: Returns a pointer to the bitmap image of the bitmap specified
//       by skinfilename. Also writes data into the bitmap header information.
//		 No support for 8-bit bitmaps.
unsigned char* cTextureInfo::LoadBitmapFile(const char *skinfilename,
	BITMAPINFOHEADER *bitmapInfoHeader)
{
	FILE *filePtr;							// the file pointer
	BITMAPFILEHEADER	bitmapFileHeader;		// bitmap file header
	
// open skinfilename in "read binary" mode
	filePtr = fopen(skinfilename, "rb");
	if (filePtr == NULL)
		return NULL;
// read the bitmap file header
	fread(&bitmapFileHeader, sizeof(BITMAPFILEHEADER), 1, filePtr);
// verify that this is a bitmap by checking for the universal bitmap id
	if (bitmapFileHeader.bfType != BITMAP_ID)
	{
		fclose(filePtr);
		return NULL;
	}
// read the bitmap information header
	fread(bitmapInfoHeader, sizeof(BITMAPINFOHEADER), 1, filePtr);
// move file pointer to beginning of bitmap data
	fseek(filePtr, bitmapFileHeader.bfOffBits, SEEK_SET);
// allocate enough memory for the bitmap image data
	unsigned int rgb_imagesize = bitmapInfoHeader->biSizeImage;
	unsigned char *rgb_image = new unsigned char[rgb_imagesize];
// allocate enough memory for the bitmap image data with alpha 
	unsigned int rgba_imagesize = 4*(rgb_imagesize/3); //Make room for alpha bytes
	unsigned char *rgba_image = new unsigned char[rgba_imagesize];
// verify memory allocation
	if (!rgb_image || !rgba_image)
	{
		delete [] rgb_image;
		delete [] rgba_image;
		fclose(filePtr);
		return NULL;
	}
// read in the bitmap image data
	fread(rgb_image, 1, bitmapInfoHeader->biSizeImage, filePtr);
// make sure bitmap image data was read
	if (rgb_image == NULL)
	{
		fclose(filePtr);
		return NULL;
	}
// swap the R and B values to get RGB since the bitmap color format is in BGR
// Also copy to the rgba_image; for now put 255 in all the alpha
	unsigned char defaultalpha = 255;
	unsigned int rgba_index = 0;
	unsigned char tempRGB;				// swap variable
	for (unsigned int rgb_index = 0; rgb_index < rgb_imagesize; rgb_index+=3)
	{
		tempRGB = rgb_image[rgb_index]; //get rvalue
		rgb_image[rgb_index] = rgb_image[rgb_index + 2]; 
			//put blue val where rvalue was
		rgb_image[rgb_index + 2] = tempRGB; //putr rvalue where blue value was
		//assert(rgba_index <= rgba_imagesize-4); //Check that sizes are ok.
		rgba_image[rgba_index] = rgb_image[rgb_index];
		rgba_index++;
		rgba_image[rgba_index] = rgb_image[rgb_index+1];
		rgba_index++;
		rgba_image[rgba_index] = rgb_image[rgb_index+2];
		rgba_index++;
		rgba_image[rgba_index] = defaultalpha;
		rgba_index++;
	}

// close the file and return the bitmap image data
	fclose(filePtr);
//	return rgb_image; 
	return rgba_image; //As of texture4.zip April 27, I use alpha in bmp files.
}


// LoadPCXTexture()
// desc: loads a PCX image file as a texture
 bool cTextureInfo::LoadPCXTexture(const char *skinfilename)
{
	PCXHEADER headerinfo;            // header of texture
	unsigned char *unscaledData;// used to calculate pcx
	int i;                             // index counter
	int j;                             // index counter
	
// load the PCX file into the texture struct
	_pdata = LoadPCXFile(skinfilename, &headerinfo);
	if (_pdata == NULL)
		return false;
// store the texture information
	_ppalette = headerinfo.palette;
	_width = headerinfo.xMax - headerinfo.xMin + 1;
	_height = headerinfo.yMax - headerinfo.yMin + 1;
	textureType = PCX;
// allocate memory for the unscaled data
	unscaledData = new unsigned char[_width*_height*4];
// store the unscaled rearranged data via the _ppalette
	for (j = 0; j < _height; j++) 
	{
		for (i = 0; i < _width; i++) 
		{
			unscaledData[4*(j*_width+i)+0] = (unsigned char)_ppalette[
				3*_pdata[j*_width+i]+0];
			unscaledData[4*(j*_width+i)+1] = (unsigned char)_ppalette[
				3*_pdata[j*_width+i]+1];
			unscaledData[4*(j*_width+i)+2] = (unsigned char)_ppalette[
				3*_pdata[j*_width+i]+2];
			unscaledData[4*(j*_width+i)+3] = (unsigned char)255;
		}
	}

// clear the texture data
	delete [] _pdata;
	_pdata = NULL;
// reallocate memory for the texture _pdata
	_pdata = new unsigned char[_width*_height*4];
// copy the rearranged unscaled data
	for (i =0; i < _width*_height*4; i++)
		_pdata[i] = unscaledData[i];
//Important, or you get memory leak.
	delete [] unscaledData;
//Done
	return true;
}


// LoadPCXFile()
// desc: returns an array of pixel data.  Also loads info into the PCXHEADER.
unsigned char*  cTextureInfo::LoadPCXFile(const char *skinfilename,
	PCXHEADER *pcxHeader)
{
	int idx = 0;                  // counter index
	int c;                        // used to retrieve a char from the file
	int i;                        // counter index
	int numRepeat;      
	FILE *filePtr;                // file handle
	int width;                    // pcx width
	int height;                   // pcx height
	unsigned char *pixelData;     // pcx image data
	unsigned char *paletteData;   // pcx palette data
	
// open PCX file
	filePtr = fopen(skinfilename, "rb");
	if (filePtr == NULL)
		return NULL;
// retrieve first character; should be equal to 10
	c = getc(filePtr);
	if (c != 10)
	{
		fclose(filePtr);
		return NULL;
	}
// retrieve next character; should be equal to 5
	c = getc(filePtr);
	if (c != 5)
	{
		fclose(filePtr);
		return NULL;
	}
// reposition file pointer to beginning of file
	rewind(filePtr);
// read 4 characters of data to skip
	fgetc(filePtr);
	fgetc(filePtr);
	fgetc(filePtr);
	fgetc(filePtr);
// retrieve leftmost x value of PCX
//	pcxHeader->xMin = fgetc(filePtr);       // loword
//	pcxHeader->xMin |= fgetc(filePtr) << 8; // hiword
	int xminlo = fgetc(filePtr);       // loword
	int xminhi = fgetc(filePtr); // hiword
	xminhi <<= 8;
	int xmin =  xminlo | xminhi;
	pcxHeader->xMin = xmin;

// retrieve bottom-most y value of PCX
	pcxHeader->yMin = fgetc(filePtr);       // loword
	pcxHeader->yMin |= fgetc(filePtr) << 8; // hiword
// retrieve rightmost x value of PCX
	pcxHeader->xMax = fgetc(filePtr);       // loword
	pcxHeader->xMax |= fgetc(filePtr) << 8; // hiword
// retrieve topmost y value of PCX
	pcxHeader->yMax = fgetc(filePtr);       // loword
	pcxHeader->yMax |= fgetc(filePtr) << 8; // hiword
// calculate the width and height of the PCX
	width = pcxHeader->xMax - pcxHeader->xMin + 1;
	height = pcxHeader->yMax - pcxHeader->yMin + 1;

// allocate memory for PCX image data
	pixelData = new unsigned char[width*height];
// set file pointer to 128th byte of file, where the PCX image data starts
	fseek(filePtr, 128, SEEK_SET);
// decode the pixel data and store
	while (idx < (width*height))
	{
		c = getc(filePtr);
		if (c > 0xbf)
		{
			numRepeat = 0x3f & c;
			c = getc(filePtr);
			for (i = 0; i < numRepeat; i++)
			{
				pixelData[idx++] = c;
			}
		}
		else
			pixelData[idx++] = c;
		fflush(stdout);
	}
// allocate memory for the PCX image palette
	paletteData = new unsigned char[768];
// palette is the last 769 bytes of the PCX file
	fseek(filePtr, -769, SEEK_END);
// verify palette; first character should be 12
	c = getc(filePtr);
	if (c != 12)
	{
		fclose(filePtr);
		return NULL;
	}
// read and store all of palette
	for (i = 0; i < 768; i++)
	{
		c = getc(filePtr);
		paletteData[i] = c;
	}
// close file and store palette in header
	fclose(filePtr);
	pcxHeader->palette = paletteData;
// return the pixel image data
	return pixelData;
}
